_ENV = AIContext

CombatAI_Method = CombatAI_Method or {}

function CombatAI_Method.init_turn()
  local new_turn = setmetatable({}, {__index = CombatAI_Method})

  new_turn:evaluate_boarding_power()

  -- lazy-populate priority caches
  new_turn.ship_priority = setmetatable({}, {
      __index = function(t, ship)
        new_turn:update_ship_priority(ship)
        return t[ship]
      end
    })
  new_turn.boarding_priority = setmetatable({}, {
      __index = function(t, ship)
        new_turn:update_boarding_priority(ship)
        return t[ship]
      end
    })
  new_turn.combat_power = setmetatable({}, {
      __index = function(t, obj)
        if is_ship(obj)
          rawset(t, obj, new_turn:ship_combat_power(obj))
        elseif is_tracking_object(obj)
          rawset(t, obj, new_turn:tracker_combat_power(obj))
        end
        return t[obj]
      end
    })

  -- find the priority of the most valuable/vulnerable friendly ship
  -- used as a hypothetical target for hostile launchers
  local current_sides_ships_proxy = setmetatable({}, { __pairs=current_sides_ships }) --because get_max_key() needs a table
  local _, max_priority = get_max_key(current_sides_ships_proxy,|) function(ship) return new_turn.ship_priority[ship] end
  new_turn.highest_friendly_priority = max_priority

  new_turn.estimated_tracker_survival, new_turn.point_def_effectiveness = evaluate_point_defences(current_side())
  print("ESTIMATED TRACKER SURVIVAL RATE", current_side().name, new_turn.estimated_tracker_survival)
  print("EST POINT DEFENCE EFFECTIVENESS", current_side().name, new_turn.point_def_effectiveness)

  new_turn:evaluate_fleet_power()
  new_turn:build_target_table()
  new_turn:build_tracker_tables() --NOTE that tracker lists may contain dead objects, due to autofire
  new_turn:build_launcher_tables()

  -- table mapping objects => list of friendly pd/interceptors reserved for that object
  new_turn.reserved_targets = setmetatable({}, { __index = function(t,k)
        rawset(t,k,{})
        return t[k]
      end
    })
  new_turn.reserved_weapons = {}

  return new_turn
end

function CombatAI_Method:update_ship_priority(target)
  self.ship_priority[target] = self:base_ship_priority(target)
end

function CombatAI_Method:update_boarding_priority(target)
  self.boarding_priority[target] = boarding_priority(target)
end

----- Target Table

function CombatAI_Method:build_target_table()

  --i.e. hostile ships
  self._hostile_targets = {}

  -- attack enemy combat ships
  for target in enemy_combat_ships() do
    self._hostile_targets[target] = true
  end

  -- if there are no combat ships left, then attack all enemy ships
  if(not next(self._hostile_targets))
    for target in current_enemy_ships() do
      self._hostile_targets[target] = true
    end
  end
end

----- Tracker Tables

function CombatAI_Method:build_tracker_tables()

  -- table of hostile trackers
  self._hostile_trackers = {}

  -- table mapping objects => list of trackers targeting that object
  -- see register_enroute_tracker()
  self.enroute_trackers = setmetatable({}, { __index = function(t,k)
        rawset(t,k,{})
        return t[k]
      end
    })

  -- populate tables
  for obj in objects() do
    if is_tracking_object(obj)
      if enemies2(obj, current_side())
        self._hostile_trackers[obj] = true
      end
      if enemies(obj, obj.dest)
        self:register_enroute_tracker(obj, obj.dest)
      end
    end
  end

end

----- Launcher Tables

function CombatAI_Method:build_launcher_tables()
  -- table of hostile launchers
  -- used when deciding to reserve interceptors for dangerous trackers not yet launched
  self._hostile_launchers = {}
  self._hostile_shuttles = {} --because assault shuttle bays are not weapons :(

  for ship in current_enemy_ships() do
    for w in weapons(ship)
      if w.is_launcher and alive_weapons(w) > 0
        self._hostile_launchers[w] = true
      end
    end

    --shuttles are not launchers, unfortunately
    local is_boarder, sys = is_combat_boarder(ship)
    if is_boarder and sys=='Assault Shuttles'
      self._hostile_shuttles[ship] = true
    end
  end
end


function CombatAI_Method:drop_target(target)
  --remove target from the target table
  self._hostile_targets[target] = nil
  self._hostile_shuttles[target] = nil

  --also drop any launchers attached to that target
  for launcher ; self._hostile_launchers do
    if launcher.object == target
      self._hostile_launchers[launcher] = nil
    end
  end
end


-- Hostile Targets Iterators

-- Since our own trackers only move at the end of our turn, it seems that we can usually rely on 
-- enemy ships dying on our turn only after we shoot at them. This is not the case for trackers, 
-- which can move on our turn and be killed by autofire (the tables regenerate at the start of our
-- turn, so we only care about things dying on our turn).
--
-- Since there doesn't seem to be any way to be notified of trackers dying due to autofire, we need
-- to check that when iterating.
--
-- Keeping the tables free of dead/captured ships OTOH is handled by calling drop_target() when appropriate.
--
function CombatAI_Method:hostile_trackers()
  return function(s,v)
    v = next(s,v)
    while v and v.dead
      v = next(s,v)
    end
    return v
  end, self._hostile_trackers
end

function CombatAI_Method:hostile_targets()
  return pairs(self._hostile_targets)
end

function CombatAI_Method:hostile_launchers()
  return pairs(self._hostile_launchers)
end

function CombatAI_Method:hostile_shuttles()
  return pairs(self._hostile_shuttles)
end


--register a tracker in flight
function CombatAI_Method:register_enroute_tracker(tracker, target)
  local tracker_data = { damage_per=average_tracker_damage(tracker), n=tracker.n }

  self.enroute_trackers[target][tracker] = tracker_data
end

--register a tracker launch. make sure this is called before launcher actually fires.
function CombatAI_Method:register_tracker_launch(launcher, target, n)
  local tracker_data = { damage_per=average_weapon_damage(launcher), ..n }

  self.enroute_trackers[target][launcher] = tracker_data
end

--register that a weapon is reserved for a potential unlaunched or autofire target
function CombatAI_Method:register_reserved_weapon(weapon, target)
  local reserved_data = { damage_per=average_weapon_damage(weapon), n=useable_weapons(weapon), ..target }

  self.reserved_targets[target][weapon] = reserved_data
  self.reserved_weapons[weapon] = reserved_data
end

function CombatAI_Method:unregister_reserved_weapon(weapon)
  local target = self.reserved_weapons[weapon].target

  self.reserved_weapons[weapon] = nil
  self.reserved_targets[target][weapon] = nil
end
