_ENV = AIContext

depends '@CombatAIMethods'
depends 'combat_helpers'
depends 'attack_priorities'
depends 'boarding'
depends 'retreat'
depends 'racial_combat_powers'

-- local we have to duplicate
local function nearest(object_iter,x,y)
  local function length(dx,dy)
    return sqrt(dx*dx+dy*dy)
  end

  local function length_to(a, b)
    ax, ay = object_center(a)
    bx, by = object_center(b)
    return length(ax-bx, ay-by)
  end

  -- local we have to duplicate
  local function object_point_dist(object,x1,y1)
    local x0,y0=object_center(object)
    return length(x0-x1,y0-y1)
  end

  local rval
  local dist
  for obj in object_iter() do
    if(not dist)
      dist=object_point_dist(obj,x,y)
      rval=obj
    else
      local mydist=object_point_dist(obj,x,y)
      if(mydist<dist)
        dist=mydist
        rval=obj
      end
    end
  end

  return rval
end

-- returns an estimate of the distance a ship might move in a single turn
local function estimated_move_dist(ship)
  if ship.cant_move
    return 0
  end

  --as a fudge factor, assume a 45 degree turn
  local turn_cost
  if ship.s < 4
    turn_cost = 25/4
  elseif ship.s == 4
    turn_cost = 25/4*0.7
  else
    turn_cost = 25/4/3
  end
  return max(ship.max_fuel - turn_cost, 0)
end

-- for debugging
local function print_target_table(ship, weapon, target_table, n)
  print(string.format("TARGETS TABLE for %s, %s", ship.name, weapon.part))
  for i, data in ipairs(target_table) do
    local name
    if data.target.is_weapon
      name = string.format("%s.%s", data.target.object.name, data.target.part)
    else
      name = data.target.name
    end
    print(string.format("%s : %.3f %s", name, data.priority, data.reserve and "(RESERVE)" or ""))
    if(n and i >= n) break end
  end
end

-- for debugging. prints out a summary of the given side's forces, 
-- including their base ship/tracker priority
local function print_force_summary(ai_data, empire)
  print("SHIPS FOR SIDE", empire.name)

  local fleet_power = 0

  print("ship : att, def, combat power : priority")
  print("----------------------------------------")
  for obj in objects() do
    if obj.empire == empire
      -- ships
      if is_ship(obj)
        local attack_value = calc_attack_value(obj) + ai_data:boarding_attack_value(obj)
        local defence_value = calc_defence_value(obj)
        local combat_power = ai_data.combat_power[obj]
        fleet_power += combat_power
        printf("%s : %.1f, %.1f, %.1f : %.3f", obj.name, attack_value, defence_value, combat_power, ai_data.ship_priority[obj])
      -- fightercraft
      elseif is_tracking_object(obj)
        local damage = average_tracker_damage(obj)
        local combat_power = ai_data.combat_power[obj]
        fleet_power += combat_power
        printf("%s : (%.1f, %.1f) x %d, %.1f : %.3f", obj.name, damage, obj.health_per, obj.n, combat_power, ai_data:base_tracker_priority_each(obj) * obj.n )
      end
    end
  end

  print("----------------------------------------")
  print("TOTAL FLEET POWER", math.round(fleet_power))
end

-- The core weapon targeting routine
function CombatAI_Method:attack_as_able(ship)
  local to_target in range_fun

  local weapons = ship.weapons
  if not weapons
    return false
  end

  local fire_map={}

  -- populate fire_map
  local j=1
  while (weapons[j]) do
    local w=weapons[j]
    j=j+1

    if self.reserved_weapons[w]
      self:unregister_reserved_weapon(w) --looks like we're re-evaluating targets for a reserved weapon
    end

    local n = w:potential_shots()

    if(n<=0) goto next_weapon end

    -- compile a table of potential targets and their priority
    local target_table = {}
    if w.is_interceptor
      -- consider reserving interceptors for trackers not yet launched
      for launcher in self:hostile_launchers() do
        target_table[#target_table+1] = { reserve=true, target=launcher, object=launcher.object, priority=self:get_launcher_priority(launcher, w) }
      end
    elseif w.is_pd
      -- consider reserving PD for trackers not yet launched, but only if they are within launch distance of us
      for launcher in self:hostile_launchers() do
        local object in launcher
        local dist = to_target(object, ship)
        local launcher_range = launcher:launcher_range() + estimated_move_dist(object)
        if dist < launcher_range
          --since we are only considering local defence here, use our own priority instead of the highest friendly priority
          local local_defence_priority = (self.ship_priority[ship] + self.highest_friendly_priority)/2
          target_table[#target_table+1] = { reserve=true, target=launcher, object=object, priority=self:get_launcher_priority(launcher, w, local_defence_priority) }
        end
      end

      -- consider reserving PD for assault shuttles
      for boarder in self:hostile_shuttles() do
        local dist = to_target(boarder, ship)
        local boarding_range = boarder:boarding_range() + estimated_move_dist(boarder)
        if(0 < dist and dist < boarding_range)
          target_table[#target_table+1] = { reserve=true, target=boarder, object=boarder, priority=self:get_assault_shuttle_priority(boarder, w) }
        end
      end
    end

    -- consider targeting trackers
    if(not w.cant_attack_trackers)
      for tracker in self:hostile_trackers() do
        if w:can_attack(tracker)
          target_table[#target_table+1] = { target=tracker, object=tracker, priority=self:get_tracker_priority(tracker, w) }
        end
      end
    end

    -- consider targeting ships
    for target in self:hostile_targets() do
      if w:can_attack(target)
        target_table[#target_table+1] = { target=target, object=target, priority=self:get_adjusted_ship_priority(target, w) }
      end
    end

    if not next(target_table)
      goto next_weapon
    end

    table.sort(target_table,|) function(a,b)
      if a.priority ~= b.priority
        return a.priority > b.priority
      end

      --prevent possible stalemate when all non-PD weapons have been destroyed
      --doing something takes precedence over doing nothing, all else equal
      return (not a.reserve) and b.reserve
    end

    --if w.is_interceptor
    --  print_target_table(ship, w, target_table)
    --end

    -- choose a target from the target table
    for i, item in ipairs(target_table) do
      local target, object, reserve in target_table[i]

      --TODO need to find a way to figure out if there are any enemy players that move after us this turn before this will work
      --[[
      -- if we are bugging out at the end of this turn, don't bother reserving anything
      if reserve and ship.retreating and ship.retreating <= 1
        goto next_target
      end
      --]]

      -- if heavy hostile PD is present, don't attack targets farther than a certain distance to limit their exposure to point-defence
      if w.is_fighter and self.estimated_tracker_survival < 1
        local max_sortie_dist = 0.45*w:launcher_range()/(1-self.estimated_tracker_survival^2)
        if reserve and is_ship(object)
          max_sortie_dist += estimated_move_dist(object)
        end
        if(launcher_dist_to_target(ship, object) > max_sortie_dist)
          goto next_target
        end
      end

      if reserve
        -- since craft don't autofire, if we are reserving and there are any targets in roundtrip range, attack those instead.
        -- We'll still be able to launch at our reserve target next turn
        if w.is_fighter
          local roundtrip_range = 0.45*w:launcher_range()
          for i, secondary in ipairs(target_table) do
            if(not secondary.reserve and launcher_dist_to_target(ship, secondary.object) < roundtrip_range)
              fire_map[#fire_map+1] = {..w,..n, target=secondary.target }
              break
            end
          end
        end

        -- weapon reserved for this turn
        self:register_reserved_weapon(w, target)
        break
      else
        -- weapon assigned to target
        fire_map[#fire_map+1] = {..w,..n,..target}
        break
      end

      ::next_target::
    end

    ::next_weapon::
  end


  -- fire weapons in batches based on whether or not they're
  -- weapon_group_equiv.
  local unfired = 0
  if fire_map[1]

    if ship.cloaked
      action.decloak {ship}
    end

    table.sort(fire_map,|) function(a,b)
        return weapon_group_order(a.w,b.w)
    end

    j=1
    for i=1,100 
      if not fire_map[j] break end
      local w,target in fire_map[j]
      local push, sub_fire = pusher {}

      while fire_map[j] and weapon_group_equiv(w,fire_map[j].w) and target==fire_map[j].target
        push | fire_map[j]
        j++
      end

      event_sep_scope(.35,action.fire_weapons, unpack(sub_fire))

      -- register the launch of any new trackers
      local w, n, target in sub_fire[1]
      if w.is_launcher
        self:register_tracker_launch(w, target, n)
      end

      -- after attacking, update the target's priority to reflect possible battle damage
      if is_ship(target)
        if target.dead
          self:drop_target(target)
        else
          self:update_ship_priority(target)
          self:update_boarding_priority(target)
        end
      end
    end
    
  end


  return #fire_map>0
end


local function invalid_move_target(ship, move_target)
  return move_target.dead or move_target.defenseless or not enemies(ship,move_target)
end

local function sort_closest(object_list,x,y)
  local function dist(ship)
    sx,sy = object_center(ship)
    return point_dist(sx,sy,x,y)
  end

  local function dist_sort(a,b)
    return dist(a) < dist(b)
  end

  table.sort(object_list, dist_sort)
end

-- chooses what to move towards
function CombatAI_Method:choose_move_target(ship, desired_attack_dist)
  local to_target in range_fun

  local closest = {}
  
  for ship in self:hostile_targets() do
    closest[#closest+1] = ship
  end

  for tracker in self:hostile_trackers() do
    closest[#closest+1] = tracker
  end
  
  sort_closest(closest, object_center(ship))

  -- choose the best move target out of the closest three ships.
  -- we don't want the AI to be too clever with how it chooses a move target here, that can actually hurt it's performance
  -- Moving the fleet towards the closest as a group turns out to help more than closing in on the 'best' target. 
  -- However, some flexibility is beneficial so we pick the best target out of the closest 3 ships
  local best_target
  local best_priority

  --print("choosing move target for", ship.name, ship.empire.name)

  local ship_count = 0
  for i, target in ipairs(closest) do
    local target_dist = to_target(ship, target)

    if is_ship(target)
      ship_count++
      if(ship_count > 3) break end
    end

    --find the best target
    local priority = 0
    if is_combat_boarder(ship)
      if is_ship(target)
        priority = self.boarding_priority[target] or 0
      end
    else
      if is_tracking_object(target)
        priority = self:base_tracker_priority_each(target) * target.n
      --planets have lowest move priority unless they are already within ~1 move of our desired attack distance
      elseif not (target.is_planet and target_dist > desired_attack_dist + estimated_move_dist(ship))
        priority = self.ship_priority[target]
      end
    end

    --print(target.name, ..target_dist, ..priority, target.dead, target.defenseless, enemies(ship,target), ..ship_count)
    if(not best_target or (priority > best_priority))
      best_target = target
      best_priority = priority
    end
  end

  return best_target
end

function CombatAI_Method:close_with_target(ship, move_target, attack_dist)
  local to_target in range_fun

  if ship.fuel > 0 and not ship.cant_move and not ship.retreating

    if stand_off_and_bombard
      attack_dist = max(attack_dist, ranges.long*1.25)
    end

    -- move and attack the current target
    local x,y,r,r2 = find_min_facing_move_L2(ship,move_target,attack_dist)

    if(x)
      if to_target(ship, move_target) <= attack_dist
        --turn to face here instead?
        x,y = ship.x, ship.y --object_center(ship)
      end

      action.move_ships{ship, x, y, r}

      if(ship.dead) return end

      if(r2 and r2~=r and ship_can_rotate(ship,r2))
        if(ship.speed==0)
          error("that's very strange")
        end
        action.move_ships{ship, x, y, r2}
      end
    end
  end

end

function CombatAI_Method:engage_targets(ship, move_target)
  local attacked = false

  -- keep calling attack_as_able until we have no more weapons to shoot
  while self:attack_as_able(ship) do
    attacked = true
    if(invalid_move_target(ship, move_target)) break end
  end

  --if the move target dies, quit attack behaviour so we can reposition against a new move target
  if(invalid_move_target(ship, move_target))
    return
  end

  if not ship.cant_electrocyte
    try_use_electrocyte(ship)

    if(invalid_move_target(ship, move_target))
      return
    end
  end

  if(not ship.has_boarded)
    -- combat boarders take more effort to choose a boarding target
    if(is_combat_boarder(ship))
      local boarding_targets = {}
      for target in self:hostile_targets() do
        if can_board(ship,target)
          boarding_targets[#boarding_targets+1] = { target=target, priority=self.boarding_priority[target] }
        end
      end

      table.sort(boarding_targets,|) function(a,b)
        return a.priority > b.priority
      end

      for i,item in ipairs(boarding_targets) do
        if(not self:capture_if_able(ship, item.target)) break end
      end

      if(not ship.has_boarded)
        self:capture_closest(ship) --fallback in case we still haven't boarded anything
      end

    -- handle opportunistic boarding by non-combat-boarders
    elseif(not attacked or all_enemies_are_immobile()) --don't attempt to board in the thick of combat
      if(self:capture_if_able(ship,move_target))
        self:capture_closest(ship)
      end
    end
  end


end

-- just fire at whatever we can, no move target
function CombatAI_Method:fire_at_will(ship)
  -- keep calling attack_as_able until we have no more weapons to shoot
  while self:attack_as_able(ship) do end
end

local function sort_move_order(ships_iter, ai_data)
  local enroute_trackers in ai_data

  local sort_ships = {}
  for ship in ships_iter() do
    sort_ships[#sort_ships+1] = {
      ..ship,
      defence_value = calc_defence_value(ship),
      interception_warning = next(enroute_trackers[ship]?) and 1 or -1,
    }
  end

  table.sort(sort_ships,|) function (b,a) --sort in reverse order, for queue
    -- first sort by interception warning, ships that have incoming trackers move last
    if a.interception_warning ~= b.interception_warning
      return a.interception_warning < b.interception_warning
    end
    -- then by defence value, so that the toughest ships to move first
    return a.defence_value > b.defence_value
  end

  local sorted = {}
  for i,ship in ipairs(sort_ships) do
    sorted[#sorted+1] = sort_ships[i].ship
  end

  return sorted
end

local function sort_attack_order(ships_iter, ai_data)
  local enroute_trackers in ai_data

  local sort_ships = {}

  for ship in current_sides_ships() do
    sort_ships[#sort_ships+1] = { 
      ..ship,
      attack_value = calc_attack_value(ship),
      is_combat_boarder = is_combat_boarder(ship) and 1 or -1,
      interception_warning = next(enroute_trackers[ship]?) and 1 or -1,
    }
  end

  table.sort(sort_ships,|) function (b,a) --sort in reverse order, for queue
    if a.interception_warning ~= b.interception_warning
      return a.interception_warning < b.interception_warning
    end

    -- combat boarders attack last
    if a.is_combat_boarder ~= b.is_combat_boarder
      return a.is_combat_boarder < b.is_combat_boarder
    end
    -- then by attack power, since high-damage ships are more likely to affect
    -- the attack decisions of later ships.
    return a.attack_value > b.attack_value
  end

  local sorted = {}
  for i,ship in ipairs(sort_ships) do
    sorted[#sorted+1] = sort_ships[i].ship
  end

  return sorted
end

-- So decoupling AI ship movement from AI ship attacks has complicated the main turn ship loop a bit
-- Instead of just iterating through all the ships in turn order, we have two separate queues.
-- Each iteration we pop a ship from the move queue, move it, then see if we can make any attacks
-- with the next ship in the attack queue.
--
-- Ships only attack after their movement has been handled, in order to give them an opportunity to
-- move within range of possible targets.
--
-- If we can attack with the next ship in the attack queue, we keep going through the attack queue until we
-- get a ship that hasn't moved yet. Then we go back into the move queue.
--
-- If a ship's move target gets destroyed, we put it back in both queues and return to handling movement.
-- This gives the ship the opportunity to move to a new target if it has movement points left
-- (and replicates the behaviour of the old close_and_attack() routine's while loop).
function CombatAI_Method:issue_ship_orders(general_retreat)
  local has_moved = {}
  local move_targets = {}

  local move_queue = sort_move_order(current_sides_ships, self)
  local attack_queue = sort_attack_order(current_sides_ships, self)

  local function retry_movement(ship)
    has_moved[ship] = nil
    move_targets[ship] = nil
    move_queue[#move_queue+1] = ship
  end

  while #move_queue > 0 and not battle_is_over() do
    -- move a ship
    local move_ship = pop_back(move_queue)

    -- if we are going to retreat this ship, we have to decide that now
    if self:wants_to_retreat(move_ship, general_retreat)
      self:do_retreat(move_ship)
    else
      local attack_dist = desired_attack_dist(move_ship, self.estimated_tracker_survival)
      if attack_dist
        local move_target = self:choose_move_target(move_ship, attack_dist) or closest_enemy_ship(object_center(move_ship))
        if move_target
          assert(not invalid_move_target(move_ship, move_target), "chose a move target that was invalid")
          move_targets[move_ship] = move_target
          self:close_with_target(move_ship, move_target, attack_dist)
        end
      end
    end

    has_moved[move_ship] = true

    -- perform attacks, if we can
    local attack_ship = attack_queue[#attack_queue]

    while #attack_queue > 0 and has_moved[attack_ship] and not battle_is_over() do
      local move_target = move_targets[attack_ship]

      if(attack_ship.dead)
        goto next_ship --looks like this ship died during movement
      end

      if move_target

        --see if our move target became invalid while we were waiting to attack
        if invalid_move_target(attack_ship, move_target)
          self:drop_target(move_target)
          retry_movement(attack_ship)
          break
        end

        self:engage_targets(attack_ship, move_target)

        --check again after attacking
        if invalid_move_target(attack_ship, move_target)
          self:drop_target(move_target)
          retry_movement(attack_ship)
          break
        end

      else
        --if cloaked, don't attack if we are retreating (since this will decloak us)
        if not (attack_ship.cloaked and attack_ship.retreating)
          self:fire_at_will(attack_ship)
        end
      end

      ::next_ship::

      --advance the attack queue
      pop_back(attack_queue)
      attack_ship = attack_queue[#attack_queue]
    end
  end

  assert(battle_is_over() or (#move_queue == 0 and #attack_queue == 0))
end

local function inside_do_turn()
  if battle_is_over()
    return
  end

  -- initialize AI data for this turn
  local ai_turn = CombatAI_Method.init_turn()
  --print_force_summary(ai_turn, current_side())

  -- do we want to retreat?
  local try_retreat_all
  if get_ai_hint() =='retreat_all' or ai_turn.outmatched > ai_turn:general_retreat_threshold()
    console("ordering a general retreat", ai_turn.outmatched)
    try_retreat_all = general_retreat()
  end

  -- stand off behaviour
  if not try_retreat_all
    --limit bombardment to the first 20 turns
    if all_enemies_are_immobile() and bombardment_available() and ai_turn.estimated_tracker_survival > 0.75 and current_side_nturns() <= 20
      print('standing off')
      stand_off_and_bombard = true
    else
      stand_off_and_bombard = false
    end
  end

  -- dispatch combat orders to ships
  ai_turn:issue_ship_orders(try_retreat_all)

  -- handle stalemate
  if (last_event_command()=="end_turn(  )"  and not stand_off_and_bombard) or current_side_nturns() > 40

    console("should probably offer a cease fire", last_event_command(), stand_off_and_bombard, current_side_nturns() )
    local empire = current_side()
    local attack = all_attacks[get_attack_id()]

    for other_empire in pairs(attack.sides) 
      if Battle.enemy(empire,other_empire) and not Battle.get_cease_fire_offer(empire,other_empire)
        console("offering a cease fire to ",other_empire,empire)
        action.offer_cease_fire(empire,other_empire)
      end

      if empire~=other_empire
        print('cease1', empire, other_empire, objects.cease_fire_offers[ 
          empire._ikey ..'_'.. other_empire._ikey ],
          Battle.get_cease_fire_offer(empire,other_empire)
        )

        print('cease2', other_empire.name, empire.name, objects.cease_fire_offers[ 
          other_empire._ikey ..'_'.. empire._ikey ],
          Battle.get_cease_fire_offer(other_empire,empire)
        )
      end
    end
  end
end


function close_and_attack_turn()
  --console("ai turn for", current_side().name, current_side(), current_side_nturns())

  --print('starting', 'temp AI State', v2s | check_var("temp AI State"))
--  console(last_event_command())

  local ok, err = pcall(inside_do_turn)
  if (not ok) and err
    add_exception(tostring | err)
  end

  action.end_turn()

end

