_ENV = AIContext

----- Boarding

-- returns true if this ship should prioritize boarding in combat
function is_combat_boarder(ship)
  local remaining_crew = ship.max_crew - ship.crew_damage
  if remaining_crew < 15
    return false
  end

  -- same check for functional assault shuttles or transporter used in boarding.lua
  local range, sys = ship:boarding_range()
  if range > 0
    return true, sys
  end

  return ship.has_boarding_module
end

local function unboardable(ship)
  return ship.drawer_table?.unboardable or ship.is_planet
end

local function estimated_boarding_crew_damage(ship)
    local boarding_crew = max(ship.max_crew - ship.crew_damage, 0)
    local crew_hitchance = action.crew_combat_strength(ship)
    return boarding_crew*(50 + crew_hitchance)/100 * 0.5 * 0.5
end


-- Return the priority of ship AS A POTENTAL TARGET FOR BOARDING.
-- boarding doesn't care about defence value, only the amount of defending crew
-- Note that the values returned from this function are NOT mutually comparable 
-- with attack priority values.
function boarding_priority(target)
  if unboardable(target)
    return 0
  end
  local remaining_crew = max(target.max_crew - target.crew_damage, 0)
  return calc_attack_value(target)/(1 + remaining_crew)
end

-- Calculate boarding hazard. This is used to evaluate how dangerous combat boarders are.
--
-- We want this to be the average defence value of our ships (since boarding bypasses defence),
-- multiplied by (estimated crew damage/defending crew). 
-- But we can't know the estimated crew damage without having the attacking ship
-- so just divide by defending crew here, and multiply by estimated casualties later (see boarding_attack_value())
local function calc_boarding_hazard(iter)
  local total = 0
  local ship_count = 0
  for obj in iter() do
    if is_ship(obj)
      if not unboardable(obj)
        local remaining_crew = max(obj.max_crew - obj.crew_damage, 0)
        local boarding_value = calc_defence_value(obj) / remaining_crew
        total += boarding_value*boarding_value --use rms so that the most highly valuable ships are weighted more
      end
      ship_count += 1
    end
  end

  return (ship_count > 0 and sqrt(total/ship_count)) or 0
end

function CombatAI_Method:evaluate_boarding_power()
  self.boarding_hazard = calc_boarding_hazard(current_sides_ships) --how threatening enemy boarders are
  self.boarding_power = calc_boarding_hazard(enemy_combat_ships)   --how valuable our boarders are
end

-- extra attack value granted to a ship based on it's boarding capabilities
function CombatAI_Method:boarding_attack_value(ship)
  local is_boarder, sys = is_combat_boarder(ship)
  if is_boarder
    local boarding_hazard = enemies2(ship, current_side()) and self.boarding_hazard or self.boarding_power
    local attack_value = boarding_hazard * estimated_boarding_crew_damage(ship)
    --print("ESTIMATED CREW DAMAGE", ship.name, estimated_boarding_crew_damage(ship))
    --if sys == 'Assault Shuttles'
    --  attack_value *= 0.85 --PD fudge factor
    --end
    return attack_value
  end

  return 0
end

-- if the target's crew is much larger than ours and
-- they are not disabled we attempt to raid them instead
local function do_boarding_action(ship, target)
  local boarders = ship.max_crew-ship.crew_damage
  local defenders = target.max_crew-target.crew_damage

  -- if we don't think we can capture the target in 3 boarding actions just raid instead
  local hack = estimated_boarding_crew_damage(ship)*2
  if(is_combat_ship(target) and ceil(defenders/hack) > 3)
    action.raid(ship,target)
  else
    action.capture(ship,target)
  end
end

-- return true if we should keep looking for another target to capture
local function capture_if_able(ship,target)
  if(not is_ship(target))
    return true
  end

  if(target.is_planet)
    return true
  end

  if(ship.dead or target.dead) 
    return true
  end

  if(ship.has_boarded)
    return false
  end

  --make sure ships don't run out of crew
  local boarders = ship.max_crew-ship.crew_damage
  local defenders = target.max_crew-target.crew_damage
  if boarders < 15 or (boarders < 30 and boarders < defenders)
    return false
  end

  if(can_board(ship,target))
    do_boarding_action(ship,target)
    return not ship.has_boarded
  end

  local x,y,r=find_max_boarding_move(ship,target,-1)

  if(x)

    if(ship.fuel==0)
      return false
    end

    action.move_ships{ship,x,y,r}

    if(ship.dead or target.dead) 
      return true
    end

    if(can_board(ship,target))
      do_boarding_action(ship,target)
    end
  end

  return not ship.has_boarded
end

-- calls capture_if_able, then drops the target if it was captured
-- otherwise ships might continue to fire on the newly captured target
function CombatAI_Method:capture_if_able(ship, target)
  local rval = capture_if_able(ship, target)
  if(not enemies(ship, target))
    self:drop_target(target)
  end
  return rval
end

function CombatAI_Method:capture_closest(ship)
  if(ship.dead) return end

  local targets=closest_objects(current_enemy_ships,object_center(ship))
  for _,target in pairs(targets) do
    if(ship.dead) return end

    if(not target.dead and not self:capture_if_able(ship,target))
      break
    end
  end
end