_ENV = AIContext

local function is_combat_ship(ship)
  for w in weapons(ship)
    if alive_weapons(w) > 0
      return true
    end
  end
  return false
end

-- don't count ships whose weapons are destroyed, as combat ships
function enemy_combat_ships()
  local f, s, v0 = current_enemy_ships()
  return function(s,v)
    local a = f(s,v)
    while a and not is_combat_ship(a)
      a=f(s,a)
    end
    return a
  end, s, v0
end

function enemies2(a,b)
  if is_object(a)
    a = a.empire
  end
  if is_object(b)
    b = b.empire
  end
  return Battle.enemy(a,b)
end

-- TODO factor in weapon mods if needed
-- average damage per mount
function average_weapon_damage(weapon)
  if weapon.is_fighter
    --while hangars bays are not counted for attack value, handling hangars here
    --is important for deciding whether or not to reserve interceptors
    local ship = weapon.object
    local fighter_weapon = get_fighter_weapon(weapon.parent_type, ship.empire, ship)
    return WEAPONS[fighter_weapon].damage_fun.average
  end

  return weapon.info.damage_fun.average
end

-- average damage per subobject
function average_tracker_damage(tracker)
  --print("TRACKER", tracker.name, tracker.is_missile, tracker.is_fighter)
  if(tracker.is_missile)
    if(tracker.submunitions > 0)
      return WEAPONS[tracker.submunitions_weapon].damage_fun.average * tracker.submunitions
    end
    return WEAPONS[tracker.warhead.parent_type].damage_fun.average
  end
  
  if(tracker.is_fighter)
    local total = 0
    --fighters only ever have 1 weapon equipped at a time right now, but loop through weapons just in case...
    for _,wname in pairs(tracker.weapons) do
      total += WEAPONS[wname.parent_type].damage_fun.average
    end
    return total
  end

  return 0
end

local function alive_and_present(w)
  local alive = alive_weapons(w)
  if w.is_fighter
    return max(0, alive - w.missing)
  end
  return alive
end

function average_tracker_evasion(speed)
  --based on the hit_chance() calculation in beam_attack_rolls.lua
  local evasion = clamp(2*speed, 0, 100)/100
  local evasion_pd = clamp(2*speed - 50, 0, 100)/100
  return (evasion + evasion_pd)/2
end

local average_attenuation = (1.0 + 0.75 + 0.25)/3 --beam_attack_rolls.lua
function default_weapon_value(w)
  -- NOTE the damage value of fightercraft are only counted when they are in their hangars
  -- it does mean that carriers drop in priority once their craft launch.
  -- that may even be a good thing, and it's certainly preferable to carriers
  -- whose craft have all been destroyed being treated as if they still had craft.
  local alive = alive_and_present(w)
  local damage = average_weapon_damage(w)

  if w.is_beam
    damage *= average_attenuation
  end

  if w.is_launcher
    --trackers with lower speed are more easily destroyed by PD
    local speed = w.info.speed or (w.is_fighter and 26 or 16) --launch_fighters.lua, launch_missiles.lua
    damage *= ( 0.1 + 0.9*average_tracker_evasion(speed) )
  end

  return damage * alive
end

-- a score that represents how "dangerous" a potential target is
function calc_attack_value(ship, weapon_value_fun)
  weapon_value_fun = weapon_value_fun or default_weapon_value

  local score = 0
  for w in weapons(ship) do
    score += weapon_value_fun(w)
  end

  return score
end


local valuable_system_types = {
  shield_regenerator = true,
  shield = true,
  capacitor = true,
  reactor = true,
  shuttle = true,
  boarding_module = true
}

local function system_value(s)
  if valuable_system_types[s.info.type]
    return 1
  end
  if s.parent_type == 'Transporter' --Transporter's system type is 'system' for some reason, so it doesn't go in valuable_system_types
    return 1
  end
  if MUNITION[s.parent_type]
    return 0.5 --killing this has a 50% chance to destroy the ship, so return 0.5. ship_damage.lua
  end
  return 0
end

-- a score that represents how "tanky" a potential target is
-- we want this to reflect how much "useless" damage we have to chew through before we start damaging systems
-- so we only include shields and armor
function calc_defence_value(ship)
  -- ignore shield and armour reflection, for simplicity

  local shield_score = 0
  if ship.adjusted_max_shields > 0
    local shield_regen = floor(ship.adjusted_max_shields*ship.shield_regen_rate)*2

    -- use rms to weight highly damaged shield segments more heavily
    -- that way if a shield segment is down it counts more than a straight average
    local rms_shield_dam = 0
    for _,s in pairs(SHIELD_NAMES)
      rms_shield_dam += ship[s]*ship[s]
    end
    rms_shield_dam = sqrt(rms_shield_dam / #SHIELD_NAMES)

    shield_score = ship.max_shields - rms_shield_dam + shield_regen
  end

  local armor_score = ship.armor - ship.armor_damage

  local useless_systems = 0
  for s in systems(ship) do
    if s.damage < s.health -- destroyed systems can't soak up damage
      useless_systems += (1-system_value(s))
    end
  end

  --is there a constant somewhere for the min structure before a ship is destroyed?
  local remaining_structure = max(ship.structure*0.7 - ship.structure_damage, 0)

  return shield_score + armor_score + min(useless_systems, remaining_structure)
end

function calc_tracker_defence_value(tracker)
  local evasion = average_tracker_evasion(tracker.speed)
  return tracker.health_per / (1 - 0.9*evasion)
end

--emulate calc_tracker_defence_value(), but for a provided hangar bay weapon
function calc_fighter_defence_value(hangar)
  local evasion = average_tracker_evasion(hangar.info.speed or 26) --launch_fighters.lua
  local health_per = hangar.info.health or 6
  return health_per / (1 - 0.9*evasion)
end

-- desired engagement ranges
local function desired_weapon_range(w, tracker_survival)
  tracker_survival = tracker_survival or 1

  if w.is_anti_missile
    return ranges.anti_missile
  end

  -- NOTE that launchers originate from their parent's center and need to reach their target's center
  -- in order to hit. So these engagement ranges need to be a bit shorter than the would be otherwise
  -- in order to compensate for that.
  if w.type=='energy_torpedo'
    return w:launcher_range() * 0.8
  end
  if w.is_fighter
    --want to be within launch range to make an attack every turn, and in case of heavy PD
    --we want to be within roundtrip range (1/2 launch range) to be back in the safety of the
    --hangars before the opponent's turn
    return w:launcher_range()*(0.3 + 0.5*tracker_survival*tracker_survival) -- 0.3 to 0.8
  end
  if w.is_launcher
    --In the case of heavy PD, prefer to be within launch range, otherwise keep a fair distance
    return w:launcher_range()*(0.3 + 4.5*tracker_survival*tracker_survival) -- 0.3 to 4.8
  end
  if w.is_pd
    return ranges.short
  end

  if w.is_beam
    --passing 1 to effective_distance() makes it return a range divisor
    return ranges.short/w:effective_distance(1)
  end
  return ranges.medium/w:effective_distance(1)
end

function desired_attack_dist(ship, tracker_survival)
  if ship.weapons?[1]?.parent_type=='Suicide Attack'
    return 3
  end

  local totalweight = 0
  local total = 0

  local totalweight_pd = 0
  local total_pd = 0

  for w in weapons(ship) do
    local range = desired_weapon_range(w, tracker_survival)

    local damage = average_weapon_damage(w) * alive_weapons(w)
    local weight = damage * damage

    if(w.is_pd)
      total_pd += weight * range
      totalweight_pd += weight
    else
      total += weight * range
      totalweight += weight
    end
  end

  --weighted average of all non-pd weapons, unless this ship only has PD weapons
  if(totalweight == 0)
    total = total_pd
    totalweight = totalweight_pd
  end

  --return nil if this ship doesn't attack
  return (totalweight > 0 and total / totalweight) or nil
end

-- since trackers need to reach their target's center, we have to use 
-- point-to-point distance to determine their range-to-target accurately
function launcher_dist_to_target(carrier, target)
  cx,cy = object_center(carrier)
  tx,ty = object_center(target)
  return point_dist(cx,cy,tx,ty)
end

----- Estimated point-defence survival rate

-- returns a survival rate for our trackers and enemy trackers
function evaluate_point_defences(side)

  -- count the total number of trackers we can launch
  -- and figure out our average tracker health
  -- TODO median tracker health might be a better metric
  local friendly_tracker_health, friendly_tracker_count = 0,0
  local hostile_tracker_health, hostile_tracker_count = 0,0

  for obj in objects() do
    if is_ship(obj)
      for w in weapons(obj) do
        if w.is_launcher
          local n = alive_weapons(w)
          if enemies2(obj, side)
            hostile_tracker_health += n * calc_fighter_defence_value(w) --defence value is "effective health", adjusted for evasion
            hostile_tracker_count += n
          else
            friendly_tracker_health += n * calc_fighter_defence_value(w)
            friendly_tracker_count += n
          end
        end
      end
    elseif is_tracking_object(obj)
      local n in obj
      if enemies2(obj, side)
        hostile_tracker_health += n * calc_tracker_defence_value(obj)
        hostile_tracker_count += n
      else
        friendly_tracker_health += n * calc_tracker_defence_value(obj)
        friendly_tracker_count += n
      end
    end
  end
  friendly_tracker_health /= friendly_tracker_count
  hostile_tracker_health /= hostile_tracker_count

  -- rough estimate of the number of trackers the enemy fleet could kill
  local friendly_point_defence = 0
  local hostile_point_defence = 0
  for obj in objects() do
    if is_ship(obj)
      local ship = obj
      for w in weapons(ship) do
        if not w.cant_attack_trackers
          --factor in PD already fired (and thus unable to autofire this turn), but not too much
          --they do get to fire them next turn, after all.
          local usable = useable_weapons(w)
          local alive = max(alive_weapons(w) - w.missing, 0)
          local n = (usable + 2*alive)/3
          local damage = average_weapon_damage(w)

          if not w.rotates
            damage *= 0.87
          end

          if enemies2(obj, side)
            local hits = min(damage/friendly_tracker_health, 1) * n
            hostile_point_defence += hits
          else
            local hits = min(damage/hostile_tracker_health, 1) * n
            friendly_point_defence += hits
          end
        end
      end
    elseif is_tracking_object(obj)
      --no real way to tell if a tracking object is an 'interceptor', hopefully this is good enough
      if obj.is_fighter and is_tracking_object(obj.dest)
        for _, wname in pairs(obj.weapons) do
          local damage = WEAPONS[wname.parent_type].damage_fun.average

          if enemies2(obj, side)
            hostile_point_defence += min(damage/friendly_tracker_health, 1) * obj.n
          else
            friendly_point_defence += min(damage/hostile_tracker_health, 1) * obj.n
          end
        end
      end
    end
  end

  --estimated tracker survival rate
  local tracker_survial_rate = friendly_tracker_count > 0 and max(friendly_tracker_count - hostile_point_defence, 0)/friendly_tracker_count or 0
  local point_defence_effectiveness = 1 - (hostile_tracker_count > 0 and max(hostile_tracker_count - friendly_point_defence, 0)/hostile_tracker_count or 0)

  return tracker_survial_rate, point_defence_effectiveness
end