_ENV = AIContext

----- Relative Fleet Combat Power Estimation

-- multiplier applied to the combat power of things that are vulnerable to point defence
local function point_defence_mult(ai_data, obj)
  local tracker_survival_rate
  if enemies2(obj, current_side())
    tracker_survival_rate = 1-ai_data.point_def_effectiveness
  else
    tracker_survival_rate = ai_data.estimated_tracker_survival
  end
  return 1 - 0.9*tracker_survival_rate
end

function CombatAI_Method:tracker_combat_power(tracker)
  local combat_power_each = average_tracker_damage(tracker)*(1+calc_tracker_defence_value(tracker)) * point_defence_mult(self, tracker)
  return combat_power_each * tracker.n
end

--emulate tracker_combat_power() but for a provided hangar bay weapon
function CombatAI_Method:fighter_combat_power(w)
  local combat_power_each = default_weapon_value(w)*(1+calc_fighter_defence_value(w)) * point_defence_mult(self, w.object)
  return combat_power_each * max(alive_weapons(w) - w.missing, 0)
end

function CombatAI_Method:ship_combat_power(ship)
  local function adjusted_weapon_value(w)
    --fightercraft can be killed without destroying their parent ship, so it makes no sense
    --to multiply their weapon value by the parent ship's defence value in the combat power calc
    if w.is_fighter
      return 0
    end
    local value = default_weapon_value(w)
    if w.is_launcher
      value *= point_defence_mult(self, w.object)
    end
    return value
  end

  local attack_value = calc_attack_value(ship, adjusted_weapon_value) + self:boarding_attack_value(ship)
  local combat_power = attack_value*(1+calc_defence_value(ship))
  for w in weapons(ship)
    if w.is_fighter
      combat_power += self:fighter_combat_power(w)
    end
  end

  if ship.cant_move
    combat_power *= 0.75 --kludge to make planets less threatening
  end

  return combat_power
end

-- used for deciding whether or not to retreat ships
function evaluate_fleet_power(ai_data)
  local allied_power = 0
  local hostile_power = 0

  for obj in objects() do
    if enemies2(obj, current_side())
      hostile_power += ai_data.combat_power[obj]
    else
      allied_power += ai_data.combat_power[obj]
    end
  end

  return allied_power, hostile_power
end

function CombatAI_Method:evaluate_fleet_power()
  local allied_power, hostile_power = evaluate_fleet_power(self)
  self.allied_combat_power = allied_power
  self.hostile_combat_power = hostile_power
  self.outmatched = (allied_power == 0 and hostile_power == 0) and 0 or hostile_power/allied_power
end

----- Retreating

-- how close are we to being destroyed?
-- return a value between 0 and 1
local function damage_assessment(ship)

  local shield = 0
  if ship.max_shields > 0
    local shield_damage = 0
    for i,s in pairs(SHIELD_NAMES) do
      shield_damage = max(shield_damage, ship[s])
    end
    shield = ship.max_shields - shield_damage
  end

  local armor = ship.armor - ship.armor_damage

  local max_structure = 0.7*ship.structure
  local structure = max(max_structure - ship.structure_damage, 0)
  
  -- since penetrating weapons do exist we don't want to rely too heavily on 
  -- undamaged shields/armor to tell us that we're not in danger
  -- so weigh them less when they are full, and more when they are damaged
  local armor_weight = ship.armor > 0 and (1 - 0.5*armor/ship.armor) or 0
  local shield_weight = ship.max_shields > 0 and (1 - 0.5*shield/ship.max_shields) or 0
  local structure_weight = max(1, armor_weight + shield_weight) --structure is always weighted at least equally to armor+shields

  local total = structure_weight*(structure/max_structure)
  total += shield_weight > 0 and shield_weight*(shield/ship.max_shields) or 0
  total += armor_weight > 0 and armor_weight*(armor/ship.armor) or 0

  local totalweight = shield_weight + armor_weight + structure_weight

  return saturate(1 - total/totalweight)
end

local function is_disarmed(ship)
  for w in weapons(ship)
   if alive_weapons(w) > 0
    return false
   end
  end
  return true
end

-- individual ship retreat decisions
function CombatAI_Method:wants_to_retreat(ship, general_retreat)
  if not can_retreat(ship)
    return false
  end

  local monster_faction in ship.empire
  if monster_faction == 'ancient_sd' or monster_faction == 'space_harpies'
    return false
  end

  -- if a general retreat has been ordered, try to retreat all ships on the same turn
  -- that means ships that can retreat faster wait until the slower ships can retreat
  if general_retreat and ship.retreat_time >= general_retreat
    return true
  end

  local severely_damaged = damage_assessment(ship) --value between 0 and 1
  --print("DAMAGE ASSESSMENT", ship.name, severely_damaged)

  if is_disarmed(ship) and (severely_damaged > 0.1 or self.outmatched > 1.5)
    print(ship.empire.name, ship.name, "retreating because disarmed")
    return true
  end

  --possibly decide to withdraw armed ships if they are really severely damaged
  --or if removing them will not swing the balance of power
  if severely_damaged > 0.5
    print(ship.empire.name, ship.name, "retreating because severely damaged", severely_damaged)
    return true
  end

  return false
end

function CombatAI_Method:do_retreat(ship)
  action.retreat(ship)
  if not ship.cloaked and not ship.cant_cloak and (is_disarmed(ship) or damage_assessment(ship) > 0.25)
    action.cloak {ship}
  end
end

function CombatAI_Method:general_retreat_threshold()
  local immobilized_power = 0

  for ship in current_sides_ships() do
    if ship.is_planet 
      return 5 --if we are defending a planet, try extra hard to hold it
    elseif ship.cant_move
      immobilized_power += self.combat_power[ship]
    end
  end

  return 3 + 2*(immobilized_power/self.allied_combat_power)
end

-- return the number of turns left until all ships retreat
function general_retreat()
  local rval = 0
  for ship in current_sides_ships() do
    if can_retreat(ship)
      if ship.retreating
        rval = max(rval, ship.retreating)
      else
        rval = max(rval, ship.retreat_time)
      end
    end
  end
  return rval
end